/*
 *
 *  VSTRING Library supporting structures and functions
 *  (c) Vladi Belperchinov-Shabanski "Cade" <cade@biscom.net> 1998-2000
 *  Distributed under the GPL license, you should receive copy of GPL!
 *
 * $Id: vstrlib.h,v 1.5 2002/04/14 10:14:44 cade Exp $
 *
 */

#ifndef _VSTRLIB_H_
#define _VSTRLIB_H_

#include <assert.h>
#ifndef ASSERT
#define ASSERT assert
#endif

#include <stdlib.h>
#include "regexp3.h"
#include "vstring.h"

/***************************************************************************
**
** GLOBALS
**
****************************************************************************/

#define VARRAY_BLOCK_SIZE 2048

class VTrie; /* forward */

/***************************************************************************
**
** VARRAY
**
****************************************************************************/

class VArray
{
  String**  _data;
  int       _size;
  int       _count;

  int       _min_len;
  int       _max_len;
  
  int       _fe;

  String    _ret_str;

  void resize( int new_size );

  void q_sort( int lo, int hi );

  public:

  VArray();
  VArray( const VArray& arr );
  VArray( const VTrie& tr );
  ~VArray();

  int count() { return _count; }
  int max_len() { return _max_len; }
  int min_len() { return _min_len; }

  void ins( int n, const char* s );
  void del( int n );
  void set( int n, const char* s );
  const char* get( int n );

  void zap();

  int push( const char* s );
  const char* pop();

  int unshift( const char* s );
  const char* shift();

  int merge( VTrie *tr ); // return new elements count
  int merge( VArray *arr ); // return new elements count

  void print();

  int fload( const char* fname ); // return 0 for ok
  int fsave( const char* fname ); // return 0 for ok
  int fload( FILE* f ); // return 0 for ok
  int fsave( FILE* f ); // return 0 for ok

  void sort();    // sort
  void reverse(); // reverse elements order
  void shuffle(); // randomize element order

  // split with regexp
  void split( const char* res, const char* str, int maxcount = -1 );
  // split with exact stringc
  void split_str( const char* res, const char* str, int maxcount = -1 );
  const char* join( const char* glue );

  String& operator []( int n )
    {
      ASSERT( n >= 0 );
      if ( n >= _count ) set( n, "" );
      return *_data[n];
    }

  const VArray& operator = ( const VArray& arr )
    { zap(); merge( (VArray*)&arr ); return *this; };
  const VArray& operator = ( const VTrie& tr )
    { zap(); merge( (VTrie*)&tr ); return *this; };
  const VArray& operator += ( const VArray& arr )
    { merge( (VArray*)&arr ); return *this; };
  const VArray& operator += ( const VTrie& tr )
    { merge( (VTrie*)&tr ); return *this; };
  
  /* utilities */
  
  /* implement `foreach'-like interface */
  void reset() // reset position to beginning
    { _fe = -1; };
  const char* next() // get next item or NULL for the end
    { _fe++; return _fe < _count ? _data[_fe]->data() : NULL; };
  const char* current() // get latest item got from next() -- current one
    { return _fe < _count ? _data[_fe]->data() : NULL; };
  int current_idx() // current index
    { return _fe; };
};

/***************************************************************************
**
** VTRIE
**
****************************************************************************/

struct VTrieNode
{
  VTrieNode() { next = down = NULL; c = 0; data = NULL; }

  VTrieNode *next;
  VTrieNode *down;
  char      c;
  String    *data;
};

class VTrie
{
  VTrieNode *root;
  int _count;
  int _depth;
  int _nodes;

  void zap_node( VTrieNode *node );
  void del_node( VTrieNode *parent, VTrieNode *node, const char* key );
  void key_node( VTrieNode *node,  VArray* arr );
  void value_node( VTrieNode *node,  VArray* arr );
  void print_node( VTrieNode* node );

  VTrieNode* find_node( const char* key );

  String temp_key; /* used for keys() */

  public:

  VTrie();
  VTrie( const VArray& arr );
  VTrie( const VTrie& tr );
  ~VTrie();

  int count() { return _count; }
  int depth() { return _depth; }
  int nodes() { return _nodes; }

  int set( const char* key, const char* data ); // return new count
  void del( const char* key );
  const char* get( const char* key );

  int exist( const char* key ) // return != 0 if key exist (with data)
      { return find_node( key ) != NULL; }

  void zap(); // delete all key+data pairs
  int keys( VArray* arr ); // return keys count
  int values( VArray* arr ); // return values count

  int merge( VTrie *tr ); // return new keys count
  int merge( VArray *arr ); // return new keys count

  void print_nodes() { print_node( root ); }
  void print();

  int fload( const char* fname ); // return 0 for ok
  int fsave( const char* fname ); // return 0 for ok
  int fload( FILE* f ); // return 0 for ok
  int fsave( FILE* f ); // return 0 for ok

  String& operator []( const char* key )
    {
    //FIXME: this is not optimal implementation...
    VTrieNode *node = find_node( key );
    if ( node ) return *(node->data);
    set( key, "" );
    node = find_node( key );
    return *(node->data);
    }

  const VTrie& operator = ( const VArray& arr )
    { zap(); merge( (VArray*)&arr ); return *this; };
  const VTrie& operator = ( const VTrie& tr )
    { zap(); merge( (VTrie*)&tr ); return *this; };
  const VTrie& operator += ( const VArray& arr )
    { merge( (VArray*)&arr ); return *this; };
  const VTrie& operator += ( const VTrie& tr )
    { merge( (VTrie*)&tr ); return *this; };
};

/***************************************************************************
**
** VREGEXP
**
****************************************************************************/

class VRegexp
{
  regexp* re;
  String substr;

  public:

  VRegexp() { re = NULL; };
  VRegexp( const char* rs ) // compiles new re
    { re = NULL; comp( rs ); };
  ~VRegexp() { if ( re ) free( re ); };

  int comp( const char* rs ) // compile re, return 1 for success
    { if ( re ) free(re); re = regcomp( rs ); return re != NULL; };
  int comp_ok() // return 1 if regexp is compiled ok, 0 if not
    { return re != NULL; }
  int exec( const char* line ) // execute re against line, return 1 for match
    { ASSERT( re ); if (!re) return 0; return regexec( re, line ); };
  int m( const char* line ) // same as exec
    { return exec( line ); };
  int m( const char* rs, const char* line ) // same as exec, but compiles first
    { comp( rs ); return m( line ); };

  const char* sub( int n ); // return n-th substring match
  String& operator []( int n ) { substr = ""; sub( n ); return substr; }
};

/***************************************************************************
**
** MISC FUNCTIONS
**
****************************************************************************/

		
		
#endif /* _VSTRLIB_H_ */


